from asynctnt import Response
from asynctnt.exceptions import TarantoolSchemaError
from tests import BaseTarantoolTestCase


class DeleteTestCase(BaseTarantoolTestCase):
    async def _fill_data(self):
        data = [
            [0, 'a', 1, 2, 'hello my darling'],
            [1, 'b', 3, 4, 'hello my darling, again'],
        ]
        for t in data:
            await self.conn.insert(self.TESTER_SPACE_ID, t)

        return data

    async def test__delete_one(self):
        data = await self._fill_data()

        res = await self.conn.delete(self.TESTER_SPACE_ID, [data[0][0]])
        self.assertIsInstance(res, Response, 'Got response')
        self.assertEqual(res.code, 0, 'success')
        self.assertGreater(res.sync, 0, 'sync > 0')
        self.assertResponseEqual(res, [data[0]], 'Body ok')

        res = await self.conn.select(self.TESTER_SPACE_ID, [0])
        self.assertResponseEqual(res, [], 'Body ok')

    async def test__delete_by_name(self):
        data = await self._fill_data()

        res = await self.conn.delete(self.TESTER_SPACE_NAME, [data[0][0]])
        self.assertIsInstance(res, Response, 'Got response')
        self.assertEqual(res.code, 0, 'success')
        self.assertGreater(res.sync, 0, 'sync > 0')
        self.assertResponseEqual(res, [data[0]], 'Body ok')

        res = await self.conn.select(self.TESTER_SPACE_ID, [0])
        self.assertResponseEqual(res, [], 'Body ok')

    async def test__delete_by_index_id(self):
        index_name = 'temp_idx'
        res = self.tnt.command(
            'make_third_index("{}")'.format(index_name)
        )
        index_id = res[0][0]

        try:
            await self.tnt_reconnect()

            data = await self._fill_data()

            res = await self.conn.delete(self.TESTER_SPACE_NAME, [data[1][2]],
                                         index=index_id)
            self.assertIsInstance(res, Response, 'Got response')
            self.assertEqual(res.code, 0, 'success')
            self.assertGreater(res.sync, 0, 'sync > 0')
            self.assertResponseEqual(res, [data[1]], 'Body ok')

            res = await self.conn.select(self.TESTER_SPACE_ID, [data[1][2]],
                                         index=index_id)
            self.assertResponseEqual(res, [], 'Body ok')
        finally:
            self.tnt.command(
                'box.space.{}.index.{}:drop()'.format(
                    self.TESTER_SPACE_NAME, index_name)
            )

    async def test__delete_by_index_name(self):
        index_name = 'temp_idx'
        res = self.tnt.command(
            'make_third_index("{}")'.format(index_name)
        )
        index_id = res[0][0]

        try:
            await self.tnt_reconnect()

            data = await self._fill_data()

            res = await self.conn.delete(self.TESTER_SPACE_NAME, [data[1][2]],
                                         index=index_name)
            self.assertIsInstance(res, Response, 'Got response')
            self.assertEqual(res.code, 0, 'success')
            self.assertGreater(res.sync, 0, 'sync > 0')
            self.assertResponseEqual(res, [data[1]], 'Body ok')

            res = await self.conn.select(self.TESTER_SPACE_ID, [data[1][2]],
                                         index=index_id)
            self.assertResponseEqual(res, [], 'Body ok')
        finally:
            self.tnt.command(
                'box.space.{}.index.{}:drop()'.format(
                    self.TESTER_SPACE_NAME, index_name)
            )

    async def test__delete_by_name_no_schema(self):
        await self.tnt_reconnect(fetch_schema=False)

        with self.assertRaises(TarantoolSchemaError):
            await self.conn.delete(self.TESTER_SPACE_NAME, [0])

    async def test__delete_by_index_name_no_schema(self):
        await self.tnt_reconnect(fetch_schema=False)

        with self.assertRaises(TarantoolSchemaError):
            await self.conn.delete(self.TESTER_SPACE_ID, [0],
                                   index='primary')

    async def test__delete_invalid_types(self):
        with self.assertRaisesRegex(
                TypeError,
                "missing 2 required positional arguments: 'space' and 'key'"):
            await self.conn.delete()

    async def test__delete_key_tuple(self):
        try:
            await self.conn.delete(self.TESTER_SPACE_ID, (1,))
        except Exception as e:
            self.fail(e)

    async def test__delete_dict_key(self):
        data = await self._fill_data()

        res = await self.conn.delete(self.TESTER_SPACE_ID, {
            'f1': 0
        })
        self.assertResponseEqual(res, [data[0]], 'Body ok')

    async def test__delete_dict_resp(self):
        data = [0, 'hello', 0, 1, 'wow']
        await self.conn.insert(self.TESTER_SPACE_ID, data)

        res = await self.conn.delete(self.TESTER_SPACE_ID, [0])
        self.assertResponseEqualKV(res, [{
            'f1': 0,
            'f2': 'hello',
            'f3': 0,
            'f4': 1,
            'f5': 'wow'
        }])



from .connection import Connection, connect
from .iproto.protocol import (
    Iterator, Response, TarantoolTuple, PushIterator,
    Schema, SchemaSpace, SchemaIndex, Metadata, Field,
    Db, IProtoError, IProtoErrorStackFrame
)

__version__ = '2.0.0'



from asynctnt import Response
from asynctnt._testbase import ensure_version
from asynctnt.prepared import PreparedStatement
from tests import BaseTarantoolTestCase


class SQLPreparedStatementTestCase(BaseTarantoolTestCase):
    @ensure_version(min=(2, 0))
    async def test__basic(self):
        stmt = self.conn.prepare('select 1, 2')
        self.assertIsInstance(stmt, PreparedStatement, 'Got correct instance')
        async with stmt:
            self.assertIsNotNone(stmt.id, 'statement has been prepared')

            res = await stmt.execute()
            self.assertIsInstance(res, Response, 'Got response')
            self.assertEqual(res.code, 0, 'success')
            self.assertGreater(res.sync, 0, 'sync > 0')
            self.assertResponseEqual(res, [[1, 2]], 'Body ok')

        self.assertIsNone(stmt.id, 'statement has been unprepared')

    @ensure_version(min=(2, 0))
    async def test__manual(self):
        stmt = self.conn.prepare('select 1, 2')
        self.assertIsInstance(stmt, PreparedStatement, 'Got correct instance')
        stmt_id = await stmt.prepare()
        self.assertIsNotNone(stmt_id, 'statement has been prepared')
        res = await stmt.execute()
        self.assertIsInstance(res, Response, 'Got response')
        self.assertEqual(res.code, 0, 'success')
        self.assertGreater(res.sync, 0, 'sync > 0')
        self.assertResponseEqual(res, [[1, 2]], 'Body ok')
        await stmt.unprepare()

    @ensure_version(min=(2, 0))
    async def test__manual_iproto(self):
        res = await self.conn.prepare_iproto('select 1, 2')
        self.assertEqual(res.code, 0, 'success')
        stmt_id = res.stmt_id
        self.assertNotEqual(stmt_id, 0, 'received statement_id')

        res = await self.conn.execute(stmt_id, [])
        self.assertIsInstance(res, Response, 'Got response')
        self.assertEqual(res.code, 0, 'success')
        self.assertGreater(res.sync, 0, 'sync > 0')
        self.assertResponseEqual(res, [[1, 2]], 'Body ok')

        res = await self.conn.unprepare_iproto(stmt_id)
        self.assertEqual(res.code, 0, 'success')

    @ensure_version(min=(2, 0))
    async def test__bind(self):
        stmt = self.conn.prepare('select 1, 2 where 1 = ? and 2 = ?')
        async with stmt:
            res = await stmt.execute([1, 2])
            self.assertIsInstance(res, Response, 'Got response')
            self.assertEqual(res.code, 0, 'success')
            self.assertGreater(res.sync, 0, 'sync > 0')
            self.assertResponseEqual(res, [[1, 2]], 'Body ok')

    @ensure_version(min=(2, 0))
    async def test__bind_metadata(self):
        stmt = self.conn.prepare('select 1, 2 where 1 = :a and 2 = :b')
        async with stmt:
            self.assertIsNotNone(stmt.params_count)
            self.assertIsNotNone(stmt.params)

            self.assertEqual(2, stmt.params_count)
            self.assertEqual(':a', stmt.params.fields[0].name)
            self.assertEqual('ANY', stmt.params.fields[0].type)
            self.assertEqual(':b', stmt.params.fields[1].name)
            self.assertEqual('ANY', stmt.params.fields[1].type)

    @ensure_version(min=(2, 0))
    async def test__bind_2_execute(self):
        stmt = self.conn.prepare('select 1, 2 where 1 = ? and 2 = ?')
        async with stmt:
            res = await stmt.execute([1, 2])
            self.assertIsInstance(res, Response, 'Got response')
            self.assertEqual(res.code, 0, 'success')
            self.assertGreater(res.sync, 0, 'sync > 0')
            self.assertResponseEqual(res, [[1, 2]], 'Body ok')

            res = await stmt.execute([3, 4])
            self.assertIsInstance(res, Response, 'Got response')
            self.assertEqual(res.code, 0, 'success')
            self.assertGreater(res.sync, 0, 'sync > 0')
            self.assertResponseEqual(res, [], 'Body is empty')

    @ensure_version(min=(2, 0))
    async def test__context_manager_double_enter(self):
        stmt = self.conn.prepare('select 1, 2 where 1 = ? and 2 = ?')
        async with stmt:
            async with stmt:  # does nothing
                res = await stmt.execute([1, 2])
                self.assertResponseEqual(res, [[1, 2]], 'Body ok')



import enum
from typing import Optional, List, Any, Union, Dict

from asynctnt.iproto import protocol

from .exceptions import TarantoolNotConnectedError
from .prepared import PreparedStatement
from .types import MethodRet, SpaceType, KeyType, TupleType


class _DbMock:
    def __getattr__(self, item):
        raise TarantoolNotConnectedError('Tarantool is not connected')


class Isolation(enum.IntEnum):
    DEFAULT = 0
    READ_COMMITTED = 1
    READ_CONFIRMED = 2
    BEST_EFFORT = 3


class Api:
    __slots__ = (
        '_db',
    )

    def __init__(self):
        self._db: Union[_DbMock, protocol.Db] = _DbMock()

    def _set_db(self, db: protocol.Db):
        self._db = db

    def _clear_db(self):
        self._db = _DbMock()

    def ping(self, *, timeout: float = -1.0) -> MethodRet:
        """
            Ping request coroutine

            :param timeout: Request timeout

            :returns: :class:`asynctnt.Response` instance
        """
        return self._db.ping(timeout=timeout)

    def call16(self,
               func_name: str,
               args: Optional[List[Any]] = None,
               *,
               timeout: float = -1.0,
               push_subscribe: bool = False) -> MethodRet:
        """
            Call16 request coroutine. It is a call with an old behaviour
            (return result of a Tarantool procedure is wrapped into a tuple,
            if needed)

            :param func_name: function name to call
            :param args: arguments to pass to the function (list object)
            :param timeout: Request timeout
            :param push_subscribe: Subscribe to push notifications

            :returns: :class:`asynctnt.Response` instance
        """
        return self._db.call16(func_name, args,
                               timeout=timeout,
                               push_subscribe=push_subscribe)

    def call(self,
             func_name: str,
             args: Optional[List[Any]] = None,
             *,
             timeout: float = -1.0,
             push_subscribe: bool = False) -> MethodRet:
        """
            Call request coroutine. It is a call with a new behaviour
            (return result of a Tarantool procedure is not wrapped into
            an extra tuple). If you're connecting to Tarantool with
            version < 1.7, then this call method acts like a call16 method

            Examples:

            .. code-block:: pycon

                # tarantool function:
                # function f(...)
                #     return ...
                # end

                >>> await conn.call('f')
                <Response sync=3 rowcount=0 data=[]>

                >>> await conn.call('f', [20, 42])
                <Response sync=3 rowcount=2 data=[20, 42]>

            :param func_name: function name to call
            :param args: arguments to pass to the function (list object)
            :param timeout: Request timeout
            :param push_subscribe: Subscribe to push notifications

            :returns: :class:`asynctnt.Response` instance
        """
        return self._db.call(func_name, args,
                             timeout=timeout, push_subscribe=push_subscribe)

    def eval(self,
             expression: str,
             args: Optional[List[Any]] = None,
             *,
             timeout: float = -1.0,
             push_subscribe: bool = False) -> MethodRet:
        """
            Eval request coroutine.

            Examples:

            .. code-block:: pycon

                >>> await conn.eval('return 42')
                <Response sync=3 rowcount=1 data=[42]>


                >>> await conn.eval('return box.info.version')
                <Response sync=3 rowcount=1 data=['2.1.1-7-gd381a45b6']>

            :param expression: expression to execute
            :param args: arguments to pass to the function, that will
                         execute your expression (list object)
            :param timeout: Request timeout
            :param push_subscribe: Subscribe to push messages

            :returns: :class:`asynctnt.Response` instance
        """
        return self._db.eval(expression, args,
                             timeout=timeout, push_subscribe=push_subscribe)

    def select(self,
               space: SpaceType,
               key: Optional[KeyType] = None,
               **kwargs) -> MethodRet:
        """
            Select request coroutine.

            Examples:

            .. code-block:: pycon

                >>> await conn.select('tester')
                <Response sync=3 rowcount=2 data=[
                    <TarantoolTuple id=1 name='one'>,
                    <TarantoolTuple id=2 name='two'>
                ]>

                >>> res = await conn.select('_space', ['tester'], index='name')
                >>> res.data
                [<TarantoolTuple id=512
                                 owner=1
                                 name='tester'
                                 engine='memtx'
                                 field_count=0
                                 flags={}
                                 format=[
                                    {'name': 'id', 'type': 'unsigned'},
                                    {'name': 'name', 'type': 'string'}
                                 ]>]


            :param space: space id or space name.
            :param key: key to select
            :param offset: offset to use
            :param limit: limit to use
            :param index: index id or name
            :param iterator: one of the following

                        * iterator id (int number),
                        * :class:`asynctnt.Iterator` object
                        * string with an iterator name

            :param timeout: Request timeout

            :returns: :class:`asynctnt.Response` instance
        """
        return self._db.select(space, key, **kwargs)

    def insert(self,
               space: SpaceType,
               t: TupleType,
               *,
               replace: bool = False,
               timeout: float = -1) -> MethodRet:
        """
            Insert request coroutine.

            Examples:

            .. code-block:: pycon

                # Basic usage
                >>> await conn.insert('tester', [0, 'hello'])
                <Response sync=3 rowcount=1 data=[
                    <TarantoolTuple id=0 name='hello'>
                ]>

                # Using dict as an argument tuple
                >>> await conn.insert('tester', {
                ...                     'id': 0
                ...                     'text': 'hell0'
                ...                   })
                <Response sync=3 rowcount=1 data=[
                    <TarantoolTuple id=0 name='hello'>
                ]>

            :param space: space id or space name.
            :param t: tuple to insert (list object)
            :param replace: performs replace request instead of insert
            :param timeout: Request timeout

            :returns: :class:`asynctnt.Response` instance
        """
        return self._db.insert(space, t,
                               replace=replace,
                               timeout=timeout)

    def replace(self,
                space: SpaceType,
                t: TupleType,
                *,
                timeout: float = -1.0) -> MethodRet:
        """
            Replace request coroutine. Same as insert, but replace.

            :param space: space id or space name.
            :param t: tuple to insert (list object)
            :param timeout: Request timeout

            :returns: :class:`asynctnt.Response` instance
        """
        return self._db.replace(space, t, timeout=timeout)

    def delete(self,
               space: SpaceType,
               key: KeyType,
               **kwargs) -> MethodRet:
        """
            Delete request coroutine.

            Examples:

            .. code-block:: pycon

                # Assuming tuple [0, 'hello'] is in space tester

                >>> await conn.delete('tester', [0])
                <Response sync=3 rowcount=1 data=[
                    <TarantoolTuple id=0 name='hello'>
                ]>

            :param space: space id or space name.
            :param key: key to delete
            :param index: index id or name
            :param timeout: Request timeout

            :returns: :class:`asynctnt.Response` instance
        """
        return self._db.delete(space, key, **kwargs)

    def update(self,
               space: SpaceType,
               key: KeyType,
               operations: List[Any],
               **kwargs) -> MethodRet:
        """
            Update request coroutine.

            Examples:

            .. code-block:: pycon

                # Assuming tuple [0, 'hello'] is in space tester

                >>> await conn.update('tester', [0], [ ['=', 1, 'hi!'] ])
                <Response sync=3 rowcount=1 data=[
                    <TarantoolTuple id=0 name='hi!'>
                ]>

                # you can use fields names as well
                >>> res = await conn.update('tester', [0],
                ...                         [ ['=', 'text', 'hola'] ])
                <Response sync=3 rowcount=1 data=[
                    <TarantoolTuple id=0 name='hola'>
                ]>

            :param space: space id or space name.
            :param key: key to update
            :param operations:
                    Operations list of the following format:
                    [ [op_type, field_no, ...], ... ]. Please refer to
                    https://tarantool.org/doc/book/box/box_space.html?highlight=update#lua-function.space_object.update
                    You can use field numbers as well as their names in space
                    format as a field_no (if only fetch_schema is True).
                    If field is unknown then TarantoolSchemaError is raised.
            :param index: index id or name
            :param timeout: Request timeout

            :returns: :class:`asynctnt.Response` instance
        """
        return self._db.update(space, key, operations, **kwargs)

    def upsert(self,
               space: SpaceType,
               t: TupleType,
               operations: List[Any],
               **kwargs) -> MethodRet:
        """
            Update request coroutine. Performs either insert or update
            (depending of either tuple exists or not)

            Examples:

            .. code-block:: pycon

                # upsert does not return anything
                >>> await conn.upsert('tester', [0, 'hello'],
                ...                   [ ['=', 1, 'hi!'] ])
                <Response sync=3 rowcount=0 data=[]>

            :param space: space id or space name.
            :param t: tuple to insert if it's not in space
            :param operations:
                    Operations list to use for update if tuple is already in
                    space. It has the same format as in update requets:
                    [ [op_type, field_no, ...], ... ]. Please refer to
                    https://tarantool.org/doc/book/box/box_space.html?highlight=update#lua-function.space_object.update
                    You can use field numbers as well as their names in space
                    format as a field_no (if only fetch_schema is True).
                    If field is unknown then TarantoolSchemaError is raised.
            :param timeout: Request timeout

            :returns: :class:`asynctnt.Response` instance
        """
        return self._db.upsert(space, t, operations, **kwargs)

    def execute(self,
                query: Union[str, int],
                args: Optional[List[Union[Dict[str, Any], Any]]] = None, *,
                parse_metadata: bool = True,
                timeout: float = -1.0) -> MethodRet:
        """
            Executes an SQL statement (only for Tarantool > 2)

            Examples:

            .. code-block:: pycon

                >>> await conn.execute("select 1 as a, 2 as b")
                <Response sync=3 rowcount=1 data=[<TarantoolTuple A=1 B=2>]>

                >>> await conn.execute("select * from sql_space")
                <Response sync=3 rowcount=2 data=[
                    <TarantoolTuple ID=1 NAME='James Bond'>,
                    <TarantoolTuple ID=2 NAME='Ethan Hunt'>
                ]>

                >>> await conn.execute("select * from sql_space",
                ...                    parse_metadata=False)
                <Response sync=3 rowcount=2 data=[
                    <TarantoolTuple 0=1 1='James Bond'>,
                    <TarantoolTuple 0=2 1='Ethan Hunt'>
                ]>

            :param query: SQL query or statement_id
            :param args: Query arguments
            :param parse_metadata: Set to False to disable response's metadata
                                   parsing for better performance
            :param timeout: Request timeout

            :returns: :class:`asynctnt.Response` instance
        """
        return self._db.execute(query, args,
                                parse_metadata=parse_metadata,
                                timeout=timeout)

    def prepare(self, query: str) -> PreparedStatement:
        """
            Create a :class:`asynctnt.prepared.PreparedStatement` instance
            :param query: query to be prepared
        """
        return PreparedStatement(self, query)

    def prepare_iproto(self,
                       query: str,
                       timeout: float = -1.0) -> MethodRet:
        """
            Low-level prepare() call
            :param query: query to be prepared
            :param timeout: request timeout
        """
        return self._db.prepare(query, timeout=timeout)

    def unprepare_iproto(self,
                         stmt_id: int,
                         timeout: float = -1.0) -> MethodRet:
        """
            Low-level unprepare() call
            :param stmt_id: query to be unprepared
            :param timeout: request timeout
        """
        return self._db.prepare(stmt_id, timeout=timeout)

    def begin(self,
              isolation: Isolation = Isolation.DEFAULT,
              tx_timeout: float = 0.0,
              timeout: float = -1.0) -> MethodRet:
        """
        Begin an interactive transaction within a stream
        :param isolation: isolation level
        :param tx_timeout: transaction timeout
        :param timeout: request timeout
        :return:
        """
        return self._db.begin(isolation.value, tx_timeout, timeout)

    def commit(self, timeout: float = -1.0) -> MethodRet:
        """
        Commit a running transaction
        :param timeout: request timeout
        """
        return self._db.commit(timeout)

    def rollback(self, timeout: float = -1.0) -> MethodRet:
        """
        Rollback a running transaction
        :param timeout: request timeout
        """
        return self._db.rollback(timeout)



import warnings

from asynctnt import TarantoolTuple
from tests import BaseTarantoolTestCase


class ResponseTestCase(BaseTarantoolTestCase):
    async def _fill_data(self, count=3):
        data = []
        for i in range(count):
            t = [i, str(i), 1, 2, 'something']
            data.append(t)
            await self.conn.insert(self.TESTER_SPACE_ID, t)
        return data

    async def _fill_data_dict(self, count=3):
        data = []
        for i in range(count):
            t = {
                'f1': i,
                'f2': str(i),
                'f3': 1,
                'f4': 2,
                'f5': 'something',
            }
            t = await self.conn.insert(self.TESTER_SPACE_ID, t)
            data.append(dict(t[0]))
        return data

    async def test__response_indexing(self):
        res = await self.conn.call('box.info')
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            self.assertEqual(len(res), len(res.body), 'len ok')
            self.assertEqual(res[0], res.body[0], 'value ok')

    async def test__response_iter(self):
        data = await self._fill_data(3)
        res = await self.conn.select(self.TESTER_SPACE_ID)

        self.assertEqual(len(res), len(data), 'len ok')

        res_arr = []
        for el in res:
            res_arr.append(list(el))
        self.assertListEqual(res_arr, data, 'list ok')

    async def test__response_tuple_iter(self):
        data = await self._fill_data(1)
        res = await self.conn.select(self.TESTER_SPACE_ID)
        t = res[0]

        t_list = [el for el in t]  # check iteration over tuple
        self.assertListEqual(t_list, data[0], 'tuple ok')

    async def test__response_tuple_keys(self):
        await self._fill_data(1)
        res = await self.conn.select(self.TESTER_SPACE_ID)
        t = res[0]

        correct_keys = ['f1', 'f2', 'f3', 'f4', 'f5']
        self.assertListEqual(list(t.keys()), correct_keys, 'keys ok')

    async def test__response_tuple_values(self):
        data = await self._fill_data(1)
        res = await self.conn.select(self.TESTER_SPACE_ID)
        t = res[0]

        self.assertListEqual(list(t.values()), data[0], 'values ok')

    async def test__response_tuple_items(self):
        data = [0, 'hello', 5, 6, 'help', 'common', 'yo']
        res = await self.conn.insert(self.TESTER_SPACE_ID, data)
        t = res[0]
        d = {
            'f1': data[0],
            'f2': data[1],
            'f3': data[2],
            'f4': data[3],
            'f5': data[4],
        }

        t_dict = {k: v for k, v in t.items()}
        self.assertDictEqual(t_dict, d, 'items ok')

    async def test__response_tuple_dict_extra_index(self):
        data = [0, 'hello', 5, 6, 'help', 'common', 'yo']
        res = await self.conn.insert(self.TESTER_SPACE_ID, data)
        res = res[0]

        self.assertEqual(res[0], data[0])
        self.assertEqual(res[1], data[1])
        self.assertEqual(res[2], data[2])
        self.assertEqual(res[3], data[3])
        self.assertEqual(res[4], data[4])
        self.assertEqual(res[5], data[5])
        self.assertEqual(res[6], data[6])
        self.assertEqual(res[-1], data[-1])
        self.assertEqual(res[-3], data[-3])

        self.assertEqual(res['f1'], data[0])
        self.assertEqual(res['f2'], data[1])
        self.assertEqual(res['f3'], data[2])
        self.assertEqual(res['f4'], data[3])
        self.assertEqual(res['f5'], data[4])

    async def test__response_tuple_slice(self):
        data = [0, 'hello', 5, 6, 'help', 'common', 'yo']
        res = await self.conn.insert(self.TESTER_SPACE_ID, data)
        res = res[0]

        self.assertEqual(type(res[:3]), tuple)

        self.assertListEqual(list(res[:3]), data[:3])
        self.assertListEqual(list(res[1:5]), data[1:5])
        self.assertListEqual(list(res[5:20]), data[5:20])
        self.assertEqual(list(res[7:3:-1]), data[7:3:-1])
        self.assertEqual(list(res[7:3:-2]), data[7:3:-2])

    async def test__response_tuple_contains(self):
        data = [0, 'hello', 5, 6, 'help', 'common', 'yo']
        res = await self.conn.insert(self.TESTER_SPACE_ID, data)
        res = res[0]

        self.assertTrue('f1' in res)
        self.assertTrue('f2' in res)
        self.assertTrue('f3' in res)
        self.assertTrue('f4' in res)
        self.assertTrue('f5' in res)
        self.assertFalse('f6' in res)

    async def test__response_tuple_key_error(self):
        data = [0, 'hello', 5, 6, 'help', 'common', 'yo']
        res = await self.conn.insert(self.TESTER_SPACE_ID, data)
        res = res[0]

        with self.assertRaises(KeyError):
            # noinspection PyStatementEffect
            res['f100']

    async def test__response_tuple_get(self):
        data = [0, 'hello', 5, 6, 'help', 'common', 'yo']
        res = await self.conn.insert(self.TESTER_SPACE_ID, data)
        res = res[0]

        self.assertEqual(res.get('f1'), 0)
        self.assertEqual(res.get('f2'), 'hello')
        self.assertEqual(res.get('f100'), None)
        self.assertEqual(res.get('f100', 'zz'), 'zz')

    async def test__response_with_no_space_format(self):
        res = await self.conn.insert('no_schema_space', [0, 'one'])

        with self.assertRaises(ValueError):
            res[0].keys()

        with self.assertRaises(ValueError):
            res[0].items()

    async def test__native_response_with_no_space_format(self):
        await self.conn.insert('no_schema_space', [0, 'one'])

        res = await self.conn.select('no_schema_space')
        self.assertEqual(1, res.rowcount, 'count correct')
        self.assertTrue(isinstance(res[0], TarantoolTuple),
                        'expecting a TarantoolTuple')
        self.assertResponseEqual(res, [[0, 'one']], 'resp ok')

        try:
            repr(res)
        except Exception as e:
            self.fail(e)

    async def test__response_repr(self):
        data = [0, 'hello', 5, 6, 'help', 'common', 'yo']
        await self.conn.insert(self.TESTER_SPACE_ID, data)

        res = await self.conn.select('tester')
        self.assertEqual(1, res.rowcount, 'count correct')
        self.assertTrue(isinstance(res[0], TarantoolTuple),
                        'expecting a TarantoolTuple')

        self.assertEqual(
            "<TarantoolTuple f1=0 f2='hello' f3=5 f4=6 f5='help' "
            "5='common' 6='yo'>",
            repr(res[0]), 'repr ok')

    async def test__response_str(self):
        data = [0, 'hello', 5, 6, 'help', 'common', 'yo']
        await self.conn.insert(self.TESTER_SPACE_ID, data)

        res = await self.conn.select('tester')
        self.assertEqual(1, res.rowcount, 'count correct')
        self.assertTrue(isinstance(res[0], TarantoolTuple),
                        'expecting a TarantoolTuple')

        try:
            str(res)
        except Exception as e:
            self.fail(e)

    async def test__metadata(self):
        self.assertIsNotNone(self.conn.schema.id)
        self.assertIsNotNone(self.conn.schema.spaces)
        self.assertIn(self.TESTER_SPACE_NAME, self.conn.schema.spaces)
        self.assertIn(self.TESTER_SPACE_ID, self.conn.schema.spaces)
        self.assertIs(self.conn.schema.spaces[self.TESTER_SPACE_NAME],
                      self.conn.schema.spaces[self.TESTER_SPACE_ID])

        sp = self.conn.schema.spaces[self.TESTER_SPACE_NAME]
        self.assertEqual(self.TESTER_SPACE_NAME, sp.name)
        self.assertEqual(self.TESTER_SPACE_ID, sp.sid)
        self.assertEqual('memtx', sp.engine)
        self.assertEqual(4, len(sp.indexes))
        self.assertIn('primary', sp.indexes)
        self.assertIn(0, sp.indexes)
        self.assertIn('txt', sp.indexes)
        self.assertIn(1, sp.indexes)
        self.assertEqual(5, len(sp.metadata.fields))
        self.assertEqual('f1', sp.metadata.fields[0].name)
        self.assertEqual('unsigned', sp.metadata.fields[0].type)
        self.assertEqual('f2', sp.metadata.fields[1].name)
        self.assertEqual('string', sp.metadata.fields[1].type)
        self.assertEqual('f5', sp.metadata.fields[4].name)
        self.assertEqual('*', sp.metadata.fields[4].type)
        self.assertEqual(5, len(sp.metadata.name_id_map))
        self.assertEqual(0, sp.metadata.name_id_map['f1'])

        idx = sp.indexes[0]
        self.assertEqual(0, idx.iid)
        self.assertEqual('primary', idx.name)
        self.assertEqual(self.TESTER_SPACE_ID, idx.sid)
        self.assertEqual('tree', idx.index_type)
        self.assertEqual(1, len(idx.metadata.fields))
        self.assertEqual('f1', idx.metadata.fields[0].name)
        self.assertEqual('unsigned', idx.metadata.fields[0].type)
        self.assertEqual(1, len(idx.metadata.name_id_map))
        self.assertEqual(0, idx.metadata.name_id_map['f1'])

    async def test__metadata_is_nullable(self):
        sp_name = "test_space_with_nullable"
        await self.conn.eval("""
            local s = box.schema.space.create('%s')
            s:format({
                {name = "id", type = "unsigned"},
                {name = "name", type = "string", is_nullable = true},
            })
        """ % (sp_name,))

        try:
            # just to be sure that schema is refreshed
            await self.conn.refetch_schema()

            self.assertIn(sp_name, self.conn.schema.spaces)
            sp = self.conn.schema.spaces[sp_name]
            self.assertEqual(sp_name, sp.name)
            self.assertEqual('memtx', sp.engine)
            self.assertEqual(0, len(sp.indexes))
            self.assertEqual(2, len(sp.metadata.fields))
            self.assertEqual('id', sp.metadata.fields[0].name)
            self.assertEqual('unsigned', sp.metadata.fields[0].type)
            self.assertEqual(None, sp.metadata.fields[0].is_nullable)
            self.assertEqual('name', sp.metadata.fields[1].name)
            self.assertEqual('string', sp.metadata.fields[1].type)
            self.assertEqual(True, sp.metadata.fields[1].is_nullable)
            self.assertEqual(2, len(sp.metadata.name_id_map))
            self.assertEqual(0, sp.metadata.name_id_map['id'])
            self.assertEqual(1, sp.metadata.name_id_map['name'])

        finally:
            await self.conn.eval("""
                local s = box.space['%s']
                if s ~= nil then
                    s:drop()
                end
            """ % (sp_name,))



from asynctnt import Response
from asynctnt._testbase import ensure_version
from tests import BaseTarantoolTestCase


class SQLExecuteTestCase(BaseTarantoolTestCase):
    @ensure_version(min=(2, 0))
    async def test__sql_basic(self):
        res = await self.conn.execute('select 1, 2')

        self.assertIsInstance(res, Response, 'Got response')
        self.assertEqual(res.code, 0, 'success')
        self.assertGreater(res.sync, 0, 'sync > 0')
        self.assertResponseEqual(res, [[1, 2]], 'Body ok')

    @ensure_version(min=(2, 0))
    async def test__sql_with_param(self):
        res = await self.conn.execute('select 1, 2 where 1 = ?', [1])

        self.assertResponseEqual(res, [[1, 2]], 'Body ok')

    @ensure_version(min=(2, 0))
    async def test__sql_with_param_cols(self):
        res = await self.conn.execute('select 1 as a, 2 as b where 1 = ?', [1])

        self.assertResponseEqualKV(res, [{'A': 1, 'B': 2}], 'Body ok')

    @ensure_version(min=(2, 0))
    async def test__sql_with_param_cols2(self):
        res = await self.conn.execute(
            'select 1 as a, 2 as b where 1 = ? and 2 = ?', [1, 2])

        self.assertResponseEqualKV(res, [{'A': 1, 'B': 2}], 'Body ok')

    @ensure_version(min=(2, 0))
    async def test__sql_with_param_cols_maps(self):
        res = await self.conn.execute(
            'select 1 as a, 2 as b where 1 = :p1 and 2 = :p2', [
                {':p1': 1},
                {':p2': 2},
            ])

        self.assertResponseEqualKV(res, [{'A': 1, 'B': 2}], 'Body ok')

    @ensure_version(min=(2, 0))
    async def test__sql_with_param_cols_maps_and_positional(self):
        res = await self.conn.execute(
            'select 1 as a, 2 as b '
            'where 1 = :p1 and 2 = :p2 and 3 = ? and 4 = ?', [
                {':p1': 1},
                {':p2': 2},
                3,
                4
            ])

        self.assertResponseEqualKV(res, [{'A': 1, 'B': 2}], 'Body ok')

    @ensure_version(min=(2, 0))
    async def test__sql_insert(self):
        res = await self.conn.execute(
            "insert into sql_space (id, name) values (1, 'one')")
        self.assertEqual(1, res.rowcount, 'rowcount ok')

    @ensure_version(min=(2, 0))
    async def test__sql_empty_autoincrement(self):
        res = await self.conn.execute(
            "insert into sql_space (id, name) values (1, 'one')")
        self.assertEqual(None, res.autoincrement_ids, 'autoincrement ok')

    @ensure_version(min=(2, 0))
    async def test__sql_insert_autoincrement(self):
        res = await self.conn.execute(
            "insert into sql_space_autoincrement (name) values ('name')")
        self.assertEqual(1, res.rowcount, 'rowcount ok')
        self.assertEqual([1], res.autoincrement_ids, 'autoincrement ok')

    @ensure_version(min=(2, 0))
    async def test__sql_insert_autoincrement_multiple(self):
        res = await self.conn.execute(
            "insert into sql_space_autoincrement_multiple (name) "
            "values ('name'), ('name2')"
        )
        self.assertEqual(2, res.rowcount, 'rowcount ok')
        self.assertEqual([1, 2], res.autoincrement_ids, 'autoincrement ok')

    @ensure_version(min=(2, 0))
    async def test__sql_insert_multiple(self):
        res = await self.conn.execute(
            "insert into sql_space (id, name) "
            "values (1, 'one'), (2, 'two')")
        self.assertEqual(2, res.rowcount, 'rowcount ok')

    @ensure_version(min=(2, 0))
    async def test__sql_update(self):
        await self.conn.execute("insert into sql_space values (1, 'one')")

        res = await self.conn.execute(
            "update sql_space set name = 'uno' where id = 1")

        self.assertEqual(1, res.rowcount, 'rowcount ok')

    @ensure_version(min=(2, 0))
    async def test__sql_update_multiple(self):
        await self.conn.execute("insert into sql_space values (1, 'one')")
        await self.conn.execute("insert into sql_space values (2, 'two')")

        res = await self.conn.execute("update sql_space set name = 'uno'")

        self.assertEqual(2, res.rowcount, 'rowcount ok')

    @ensure_version(min=(2, 0))
    async def test__sql_delete(self):
        await self.conn.execute("insert into sql_space values (1, 'one')")
        res = await self.conn.execute(
            "delete from sql_space where name = 'one'"
        )
        self.assertEqual(1, res.rowcount, 'rowcount ok')

    @ensure_version(min=(2, 0))
    async def test__sql_select(self):
        await self.conn.execute("insert into sql_space values (1, 'one')")
        await self.conn.execute("insert into sql_space values (2, 'two')")

        res = await self.conn.execute("select * from sql_space")
        self.assertEqual(2, res.rowcount, 'rowcount is surely ok')
        self.assertEqual(1, res.body[0]['ID'])
        self.assertEqual('one', res.body[0]['NAME'])
        self.assertEqual(2, res.body[1]['ID'])
        self.assertEqual('two', res.body[1]['NAME'])

    @ensure_version(min=(2, 0))
    async def test__sql_delete_multiple(self):
        await self.conn.execute("insert into sql_space values (1, 'one')")
        await self.conn.execute("insert into sql_space values (2, 'two')")

        res = await self.conn.execute("delete from sql_space")
        self.assertEqual(2, res.rowcount, 'rowcount ok')

        res = await self.conn.execute("select * from sql_space")
        self.assertEqual(0, res.rowcount, 'rowcount is surely ok')

    @ensure_version(min=(2, 0))
    async def test__metadata(self):
        res = await self.conn.execute('select 1, 2')
        self.assertIsNotNone(res.metadata)
        self.assertIsNotNone(res.metadata.fields)
        self.assertEqual(2, len(res.metadata.fields))

        self.assertEqual('COLUMN_1', res.metadata.fields[0].name)
        self.assertEqual('integer', res.metadata.fields[0].type)
        self.assertEqual('COLUMN_2', res.metadata.fields[1].name)
        self.assertEqual('integer', res.metadata.fields[1].type)

    @ensure_version(min=(2, 0))
    async def test__metadata_names(self):
        res = await self.conn.execute('select 1 as a, 2 as b')
        self.assertIsNotNone(res.metadata)
        self.assertIsNotNone(res.metadata.fields)
        self.assertEqual(2, len(res.metadata.fields))

        self.assertEqual('A', res.metadata.fields[0].name)
        self.assertEqual('integer', res.metadata.fields[0].type)
        self.assertEqual('B', res.metadata.fields[1].name)
        self.assertEqual('integer', res.metadata.fields[1].type)

    @ensure_version(min=(2, 0))
    async def test__metadata_actual_space(self):
        await self.conn.execute("insert into sql_space values (1, 'one')")
        await self.conn.execute("insert into sql_space values (2, 'two')")

        res = await self.conn.execute("select * from sql_space")
        self.assertEqual(2, res.rowcount, 'rowcount is ok')
        self.assertEqual(2, len(res.metadata.fields))
        self.assertEqual('ID', res.metadata.fields[0].name)
        self.assertEqual('integer', res.metadata.fields[0].type)
        self.assertIsNone(res.metadata.fields[0].is_nullable)
        self.assertIsNone(res.metadata.fields[0].is_autoincrement)
        self.assertIsNone(res.metadata.fields[0].collation)

        self.assertEqual('NAME', res.metadata.fields[1].name)
        self.assertEqual('string', res.metadata.fields[1].type)
        self.assertIsNone(res.metadata.fields[1].is_nullable)
        self.assertIsNone(res.metadata.fields[1].is_autoincrement)
        self.assertIsNone(res.metadata.fields[1].collation)

    @ensure_version(min=(2, 0))
    async def test__sql_select_full_metadata(self):
        await self.conn.execute("insert into sql_space values (1, 'one')")
        await self.conn.execute("insert into sql_space values (2, 'two')")

        await self.conn.update(
            '_session_settings',
            ['sql_full_metadata'],
            [('=', 'value', True)]
        )

        try:
            res = await self.conn.execute("select * from sql_space")
            self.assertEqual(2, len(res.metadata.fields))
            self.assertEqual('ID', res.metadata.fields[0].name)
            self.assertEqual('integer', res.metadata.fields[0].type)
            self.assertEqual(False, res.metadata.fields[0].is_nullable)
            self.assertEqual(None, res.metadata.fields[1].is_autoincrement)
            self.assertIsNone(res.metadata.fields[0].collation)

            self.assertEqual('NAME', res.metadata.fields[1].name)
            self.assertEqual('string', res.metadata.fields[1].type)
            self.assertEqual(True, res.metadata.fields[1].is_nullable)
            self.assertEqual(None, res.metadata.fields[1].is_autoincrement)
            self.assertEqual('unicode', res.metadata.fields[1].collation)
        finally:
            await self.conn.update(
                '_session_settings',
                ['sql_full_metadata'],
                [('=', 'value', False)]
            )



import asyncio
import enum
import functools
import os
from typing import Optional, Union

from .api import Api
from .exceptions import TarantoolDatabaseError, \
    ErrorCode, TarantoolError
from .iproto import protocol
from .log import logger
from .stream import Stream
from .utils import get_running_loop

__all__ = (
    'Connection', 'connect', 'ConnectionState'
)


class ConnectionState(enum.IntEnum):
    CONNECTING = 1
    CONNECTED = 2
    RECONNECTING = 3
    DISCONNECTING = 4
    DISCONNECTED = 5


class Connection(Api):
    __slots__ = (
        '_host', '_port', '_username', '_password',
        '_fetch_schema', '_auto_refetch_schema', '_initial_read_buffer_size',
        '_encoding', '_connect_timeout', '_reconnect_timeout',
        '_request_timeout', '_ping_timeout', '_loop', '_state', '_state_prev',
        '_transport', '_protocol',
        '_disconnect_waiter', '_reconnect_task',
        '_connect_lock', '_disconnect_lock',
        '_ping_task', '__create_task'
    )

    def __init__(self, *,
                 host: str = '127.0.0.1',
                 port: Union[int, str] = 3301,
                 username: Optional[str] = None,
                 password: Optional[str] = None,
                 fetch_schema: bool = True,
                 auto_refetch_schema: bool = True,
                 connect_timeout: float = 3.,
                 request_timeout: float = -1.,
                 reconnect_timeout: float = 1. / 3.,
                 ping_timeout: float = 5.,
                 encoding: Optional[str] = None,
                 initial_read_buffer_size: Optional[int] = None):

        """
            Connection constructor.

            To manipulate a Connection instance there are several functions:

                * await connect() - performs connecting, authorization and
                                    schema fetching.

                * await disconnect() - performs disconnection.

                * close() - closes connection (not a coroutine)

            Connection also supports context manager protocol, which connects
            on entering and disconnecting on leaving a block.

            So one can simply use it as follows:

            .. code-block:: python

                async with asynctnt.Connection() as conn:
                    await conn.call('box.info')

            :param host:
                    Tarantool host (pass ``unix/`` to connect to unix socket)
            :param port:
                    Tarantool port
                    (pass ``/path/to/sockfile`` to connect ot unix socket)
            :param username:
                    Username to use for auth
                    (if ``None`` you are connected as a guest)
            :param password:
                    Password to use for auth
            :param fetch_schema:
                    Pass ``True`` to be able to use spaces and indexes names in
                    data manipulation routines (default is ``True``)
            :param auto_refetch_schema:
                    If set to ``True`` then when ER_WRONG_SCHEMA_VERSION error
                    occurs on a request, schema is refetched and the initial
                    request is resent. If set to ``False`` then schema will not
                    be checked by Tarantool, so no errors will occur
            :param connect_timeout:
                    Time in seconds how long to wait for connecting to socket
            :param request_timeout:
                    Request timeout (in seconds) for all requests
                    (by default there is no timeout)
            :param reconnect_timeout:
                    Time in seconds to wait before automatic reconnect
                    (set to ``0`` or ``None`` to disable auto reconnect)
            :param ping_timeout:
                    If specified (default is 5 seconds) a background task
                    will be created which will ping Tarantool instance
                    periodically to check if it is alive and update schema
                    if it is changed
                    (set to ``0`` or ``None`` to disable this task)
            :param encoding:
                    The encoding to use for all strings
                    encoding and decoding (default is ``utf-8``)
            :param initial_read_buffer_size:
                    Initial and minimum size of read buffer in bytes.
                    Higher value means less reallocations, but higher
                    memory usage (default is 131072).
        """
        super().__init__()
        self._host = host
        self._port = port
        self._username = username
        self._password = password
        self._fetch_schema = False if fetch_schema is None else fetch_schema
        if auto_refetch_schema:  # None hack
            self._auto_refetch_schema = True
            if not self._fetch_schema:
                logger.warning('Setting fetch_schema to True as '
                               'auto_refetch_schema is True')
                self._fetch_schema = True
        else:
            self._auto_refetch_schema = False
        self._initial_read_buffer_size = initial_read_buffer_size
        self._encoding = encoding or 'utf-8'

        self._connect_timeout = connect_timeout
        self._reconnect_timeout = reconnect_timeout or 0
        self._request_timeout = request_timeout
        self._ping_timeout = ping_timeout or 0

        self._loop = None
        self.__create_task = None

        self._transport = None
        self._protocol: Optional[protocol.Protocol] = None

        self._state = ConnectionState.DISCONNECTED
        self._state_prev = ConnectionState.DISCONNECTED
        self._disconnect_waiter = None
        self._reconnect_task = None
        self._connect_lock = asyncio.Lock()
        self._disconnect_lock = asyncio.Lock()
        self._ping_task = None

    def _set_state(self, new_state: ConnectionState):
        if self._state != new_state:
            logger.debug('Changing state %s -> %s',
                         self._state.name, new_state.name)
            self._state_prev = self._state
            self._state = new_state

    def connection_lost(self, exc):
        if self._transport:
            self._transport.close()
        self._transport = None

        if self._disconnect_waiter:
            # disconnect() call happened
            self._disconnect_waiter.set_result(True)
            return

        # connection lost

        if self._reconnect_timeout > 0:
            # should reconnect
            self._start_reconnect(return_exceptions=False)
        else:
            # should not reconnect, close everything
            self.close()

    async def _ping_task_func(self):
        while self._state == ConnectionState.CONNECTED:
            try:
                await self.ping(timeout=2.0)
            except asyncio.CancelledError:
                break
            except Exception:
                pass

            await asyncio.sleep(self._ping_timeout)

    def _start_reconnect(self, return_exceptions: bool = False):
        if self._state in [ConnectionState.CONNECTING,
                           ConnectionState.RECONNECTING]:
            logger.debug('%s Cannot start reconnect: already reconnecting',
                         self.fingerprint)
            return

        if self._reconnect_task:  # pragma: nocover
            return

        logger.info('%s Started reconnecting', self.fingerprint)
        self._set_state(ConnectionState.RECONNECTING)

        self._reconnect_task = self.__create_task(
            self._connect(return_exceptions=return_exceptions)
        )

    def protocol_factory(self,
                         connected_fut: asyncio.Future,
                         cls=protocol.Protocol):
        return cls(host=self._host,
                   port=self._port,
                   username=self._username,
                   password=self._password,
                   fetch_schema=self._fetch_schema,
                   auto_refetch_schema=self._auto_refetch_schema,
                   request_timeout=self._request_timeout,
                   initial_read_buffer_size=self._initial_read_buffer_size,
                   encoding=self._encoding,
                   connected_fut=connected_fut,
                   on_connection_made=None,
                   on_connection_lost=self.connection_lost,
                   loop=self._loop)

    async def _connect(self, return_exceptions: bool = True):
        if self._loop is None:
            self._loop = get_running_loop()
            if hasattr(self._loop, 'create_task'):
                self.__create_task = self._loop.create_task
            else:  # pragma: nocover
                self.__create_task = asyncio.ensure_future

        async with self._connect_lock:
            while True:
                try:
                    ignore_states = {
                        ConnectionState.CONNECTED,
                        ConnectionState.DISCONNECTING,  # disconnect() called
                    }

                    if self._state in ignore_states:
                        self._reconnect_task = None
                        return

                    self._set_state(ConnectionState.CONNECTING)

                    async def full_connect():
                        while True:
                            connected_fut = _create_future(self._loop)

                            if self._host.startswith('unix/'):
                                unix_path = self._port
                                assert isinstance(unix_path, str), \
                                    'port must be a str instance for ' \
                                    'unix socket'
                                assert unix_path, \
                                    'No unix file path specified'
                                assert os.path.exists(unix_path), \
                                    'Unix socket `{}` not found'.format(
                                        unix_path)

                                conn = self._loop.create_unix_connection(
                                    functools.partial(self.protocol_factory,
                                                      connected_fut),
                                    unix_path
                                )
                            else:
                                conn = self._loop.create_connection(
                                    functools.partial(self.protocol_factory,
                                                      connected_fut),
                                    self._host, self._port)

                            tr, pr = await conn

                            try:
                                timeout = 0.05  # wait at least something
                                if self._connect_timeout is not None:
                                    timeout = self._connect_timeout / 2

                                await asyncio.wait_for(connected_fut,
                                                       timeout=timeout)
                            except asyncio.TimeoutError:  # pragma: nocover
                                tr.close()
                                continue  # try again
                            except Exception:
                                tr.close()
                                raise

                            return tr, pr

                    tr, pr = await asyncio.wait_for(
                        full_connect(),
                        timeout=self._connect_timeout,
                    )

                    logger.info('%s Connected successfully', self.fingerprint)
                    self._set_state(ConnectionState.CONNECTED)

                    self._transport = tr
                    self._protocol = pr
                    self._set_db(self._protocol.get_common_db())
                    self._reconnect_task = None
                    self._normalize_api()

                    if self._ping_timeout:
                        self._ping_task = \
                            self.__create_task(self._ping_task_func())
                    return
                except TarantoolDatabaseError as e:
                    skip_errors = {
                        ErrorCode.ER_LOADING,
                        ErrorCode.ER_NO_SUCH_SPACE,
                        ErrorCode.ER_NO_SUCH_INDEX_ID
                    }
                    if e.code in skip_errors:
                        # If Tarantool is still loading then reconnect
                        if self._reconnect_timeout > 0:
                            await self._wait_reconnect(e)
                            continue

                    if return_exceptions:
                        self._reconnect_task = None
                        raise e

                    logger.exception(e)
                    if self._reconnect_timeout > 0:
                        await self._wait_reconnect(e)
                        continue

                    return  # no reconnect, no return_exceptions
                except asyncio.CancelledError:
                    logger.debug("connect is cancelled")
                    self._reconnect_task = None
                    raise
                except Exception as e:
                    if self._reconnect_timeout > 0:
                        await self._wait_reconnect(e)
                        continue

                    if return_exceptions:
                        self._reconnect_task = None
                        raise e

                    logger.exception(e)
                    return  # no reconnect, no return_exceptions

    async def _wait_reconnect(self, exc=None):
        self._set_state(ConnectionState.RECONNECTING)
        logger.warning('Connect to %s failed: %s. Retrying in %f seconds',
                       self.fingerprint,
                       repr(exc) if exc else '',
                       self._reconnect_timeout)

        await asyncio.sleep(self._reconnect_timeout)

    async def connect(self) -> 'Connection':
        """
            Connect coroutine
        """
        await self._connect(True)
        return self

    async def disconnect(self):
        """
            Disconnect coroutine
        """

        async with self._disconnect_lock:
            if self._state == ConnectionState.DISCONNECTED:
                return

            self._set_state(ConnectionState.DISCONNECTING)

            logger.info('%s Disconnecting...', self.fingerprint)
            if self._reconnect_task:
                self._reconnect_task.cancel()
                self._reconnect_task = None

            if self._ping_task and not self._ping_task.done():
                self._ping_task.cancel()
                self._ping_task = None

            self._clear_db()
            if self._transport:
                self._disconnect_waiter = _create_future(self._loop)
                self._transport.close()
                self._transport = None
                self._protocol = None

                await self._disconnect_waiter
                self._disconnect_waiter = None
                self._set_state(ConnectionState.DISCONNECTED)
            else:
                self._transport = None
                self._protocol = None
                self._disconnect_waiter = None
                self._set_state(ConnectionState.DISCONNECTED)

    def close(self):
        """
            Same as disconnect, but not a coroutine, i.e. it does not wait
            for disconnect to finish.
        """

        if self._state == ConnectionState.DISCONNECTED:
            return

        self._set_state(ConnectionState.DISCONNECTING)
        logger.info('%s Disconnecting...', self.fingerprint)

        if self._reconnect_task and not self._reconnect_task.done():
            self._reconnect_task.cancel()
            self._reconnect_task = None

        if self._ping_task and not self._ping_task.done():
            self._ping_task.cancel()
            self._ping_task = None

        if self._transport:
            self._transport.close()

        self._transport = None
        self._protocol = None
        self._disconnect_waiter = None
        self._clear_db()
        self._set_state(ConnectionState.DISCONNECTED)

    async def reconnect(self):
        """
            Reconnect coroutine.
            Just calls disconnect() and connect()
        """
        await self.disconnect()
        await self.connect()

    async def __aenter__(self):
        """
            Executed on entering the async with section.
            Connects to Tarantool instance.
        """
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """
            Executed on leaving the async with section.
            Disconnects from Tarantool instance.
        """
        await self.disconnect()

    @property
    def fingerprint(self) -> str:
        return 'Tarantool[{}:{}]'.format(self._host, self._port)

    @property
    def host(self) -> str:
        """
            Tarantool host
        """
        return self._host

    @property
    def port(self) -> int:
        """
            Tarantool port
        """
        return self._port

    @property
    def username(self) -> Optional[str]:
        """
            Tarantool username
        """
        return self._username

    @property
    def password(self) -> Optional[str]:
        """
            Tarantool password
        """
        return self._password

    @property
    def fetch_schema(self) -> bool:
        """
            fetch_schema flag
        """
        return self._fetch_schema

    @property
    def auto_refetch_schema(self) -> bool:
        """
            auto_refetch_schema flag
        """
        return self._auto_refetch_schema

    @property
    def encoding(self) -> str:
        """
            Connection encoding
        """
        return self._encoding

    @property
    def reconnect_timeout(self) -> float:
        """
            Reconnect timeout value
        """
        return self._reconnect_timeout

    @property
    def connect_timeout(self) -> float:
        """
            Connect timeout value
        """
        return self._connect_timeout

    @property
    def request_timeout(self) -> float:
        """
            Request timeout value
        """
        return self._request_timeout

    @property
    def version(self) -> Optional[tuple]:
        """
            Protocol version tuple. ex.: (1, 6, 7)
        """
        if self._protocol is None:
            return None
        return self._protocol.get_version()

    @property
    def loop(self):
        """
            Asyncio event loop
        """
        return self._loop

    @property
    def state(self) -> ConnectionState:
        """
            Current connection state

            :rtype: ConnectionState
        """
        return self._state

    @property
    def is_connected(self) -> bool:
        """
            Check if an underlying connection is active
        """
        if self._protocol is None:
            return False
        return self._protocol.is_connected()

    @property
    def is_fully_connected(self) -> bool:
        """
            Check if connection is fully active (performed auth
            and schema fetching)
        """
        if self._protocol is None:
            return False
        return self._protocol.is_fully_connected()

    @property
    def schema_id(self) -> Optional[int]:
        """
            Tarantool's current schema id
        """
        if self._protocol is None:
            return None
        return self._protocol.schema_id

    @property
    def schema(self) -> Optional[protocol.Schema]:
        """
            Current Tarantool schema with all spaces, indexes and fields
        """
        if self._protocol is None:  # pragma: nocover
            return None
        return self._protocol.schema

    @property
    def initial_read_buffer_size(self) -> int:
        """
            initial_read_buffer_size value
        """
        return self._initial_read_buffer_size

    async def refetch_schema(self):
        """
            Coroutine to force refetch schema
        """
        await self._protocol.refetch_schema()

    def _normalize_api(self):
        if (1, 6) <= self.version < (1, 7):  # pragma: nocover
            Api.call = Api.call16
            Connection.call = Connection.call16

        if self.version < (2, 10):  # pragma: nocover
            def stream_stub(_):
                raise TarantoolError(
                    "streams are available only in Tarantool 2.10+"
                )

            Connection.stream = stream_stub

    def __repr__(self):
        return "<asynctnt.Connection host={} port={} state={}>".format(
            self.host,
            self.port,
            repr(self.state)
        )

    def stream(self) -> Stream:
        """
            Create new stream suitable for interactive transactions
        """
        stream = Stream()
        db = self._protocol.create_db(True)
        stream._set_db(db)
        return stream


def _create_future(loop):
    try:
        return loop.create_future()
    except AttributeError:  # pragma: nocover
        return asyncio.Future(loop=loop)


async def connect(**kwargs) -> Connection:
    """
        connect shorthand. See :class:`asynctnt.Connection` for kwargs details

        :return: :class:`asynctnt.Connection` object
    """
    return await Connection(**kwargs).connect()



import asyncio
import contextlib
import functools
import inspect
import logging
import os
import time
import unittest

import sys
from typing import Tuple, Optional

import asynctnt
from asynctnt.instance import \
    TarantoolSyncInstance, TarantoolSyncDockerInstance
from asynctnt import TarantoolTuple
from asynctnt.utils import get_running_loop

__all__ = (
    'TestCase', 'TarantoolTestCase', 'ensure_version', 'check_version',
    'ensure_bin_version',
)


@contextlib.contextmanager
def silence_asyncio_long_exec_warning():
    def flt(log_record):
        msg = log_record.getMessage()
        return not msg.startswith('Executing ')

    logger = logging.getLogger('asyncio')
    logger.addFilter(flt)
    try:
        yield
    finally:
        logger.removeFilter(flt)


class TestCaseMeta(type(unittest.TestCase)):

    @staticmethod
    def _iter_methods(bases, ns):
        for base in bases:
            for methname in dir(base):
                if not methname.startswith('test_'):
                    continue

                meth = getattr(base, methname)
                if not inspect.iscoroutinefunction(meth):
                    continue

                yield methname, meth

        for methname, meth in ns.items():
            if not methname.startswith('test_'):
                continue

            if not inspect.iscoroutinefunction(meth):
                continue

            yield methname, meth

    def __new__(mcls, name, bases, ns):
        for methname, meth in mcls._iter_methods(bases, ns):
            @functools.wraps(meth)
            def wrapper(self, *args, __meth__=meth, **kwargs):
                self.loop.run_until_complete(__meth__(self, *args, **kwargs))

            ns[methname] = wrapper

        return super().__new__(mcls, name, bases, ns)


class TestCase(unittest.TestCase, metaclass=TestCaseMeta):
    _loop = None
    _use_uvloop = False

    @classmethod
    def setUpClass(cls):
        cls._use_uvloop = os.environ.get('USE_UVLOOP')

        if cls._use_uvloop:
            import uvloop
            uvloop.install()

        try:
            loop = get_running_loop()
            # if there is a running loop - close it
            loop.close()
        except RuntimeError:
            pass

        cls._loop = asyncio.new_event_loop()
        asyncio.set_event_loop(cls._loop)

        if cls._use_uvloop:
            import uvloop
            assert isinstance(cls._loop, uvloop.Loop)

    @classmethod
    def tearDownClass(cls):
        if cls._loop is not None:
            cls._loop.close()
            cls._loop = None

    @property
    def loop(self):
        return self._loop

    @contextlib.contextmanager
    def assertRunUnder(self, delta):
        st = time.monotonic()
        try:
            yield
        finally:
            if time.monotonic() - st > delta:
                raise AssertionError(
                    'running block took longer than {}'.format(delta))

    @classmethod
    def ensure_future(cls, coro_or_future):
        return asyncio.ensure_future(coro_or_future)

    @classmethod
    def sleep(cls, delay, result=None):
        return asyncio.sleep(delay, result)


class TarantoolTestCase(TestCase):
    DO_CONNECT = True
    LOGGING_LEVEL = logging.WARNING
    LOGGING_STREAM = sys.stderr
    TNT_APP_LUA_PATH = None
    TNT_CLEANUP = True
    EXTRA_BOX_CFG = ""

    tnt = None
    in_docker = False

    @classmethod
    def read_applua(cls):
        if cls.TNT_APP_LUA_PATH:
            with open(cls.TNT_APP_LUA_PATH, 'r') as f:
                return f.read()

    @classmethod
    def init_instance(cls):
        tnt, in_docker = cls._make_instance(
            extra_box_cfg=cls.EXTRA_BOX_CFG
        )
        cls.tnt = tnt
        cls.in_docker = in_docker

    @classmethod
    def setUpClass(cls):
        TestCase.setUpClass()
        logging.basicConfig(level=cls.LOGGING_LEVEL,
                            stream=cls.LOGGING_STREAM)
        if cls.tnt is None:
            tnt, in_docker = cls._make_instance(
                extra_box_cfg=cls.EXTRA_BOX_CFG
            )
            cls.tnt = tnt
            cls.in_docker = in_docker

        cls.tnt.start()

    @classmethod
    def make_instance(cls):
        obj, _ = cls._make_instance()
        return obj

    @classmethod
    def _make_instance(cls, **kwargs):
        tarantool_docker_image = os.getenv('TARANTOOL_DOCKER_IMAGE')
        tarantool_docker_tag = os.getenv('TARANTOOL_DOCKER_VERSION')
        in_docker = False
        if tarantool_docker_tag:
            print('Running tarantool in docker: {}:{}'.format(
                tarantool_docker_image or 'tarantool/tarantool',
                tarantool_docker_tag))

            tnt = TarantoolSyncDockerInstance(
                applua=cls.read_applua(),
                docker_image=tarantool_docker_image,
                docker_tag=tarantool_docker_tag,
                timeout=4 * 60,
                **kwargs
            )
            in_docker = True
        else:
            unix_path = os.getenv('TARANTOOL_LISTEN_UNIX_PATH')
            if not unix_path:
                tnt = TarantoolSyncInstance(
                    port=TarantoolSyncInstance.get_random_port(),
                    console_port=TarantoolSyncInstance.get_random_port(),
                    applua=cls.read_applua(),
                    cleanup=cls.TNT_CLEANUP,
                    **kwargs
                )
            else:
                tnt = TarantoolSyncInstance(
                    host='unix/',
                    port=unix_path,
                    console_host='127.0.0.1',
                    applua=cls.read_applua(),
                    cleanup=cls.TNT_CLEANUP,
                    **kwargs
                )

        return tnt, in_docker

    @classmethod
    def tearDownClass(cls):
        if cls.tnt:
            cls.tnt.stop()
        TestCase.tearDownClass()

    def setUp(self):
        super(TarantoolTestCase, self).setUp()
        if self.DO_CONNECT:
            self.loop.run_until_complete(self.tnt_connect())

    def tearDown(self):
        self.loop.run_until_complete(self.tnt_disconnect())
        super(TarantoolTestCase, self).tearDown()

    @property
    def conn(self) -> asynctnt.Connection:
        return self._conn

    async def tnt_connect(self, *,
                          username=None, password=None,
                          fetch_schema=True,
                          auto_refetch_schema=False,
                          connect_timeout=None, reconnect_timeout=1 / 3,
                          ping_timeout=0,
                          request_timeout=None, encoding='utf-8',
                          initial_read_buffer_size=None):
        self._conn = asynctnt.Connection(
            host=self.tnt.host,
            port=self.tnt.port,
            username=username,
            password=password,
            fetch_schema=fetch_schema,
            auto_refetch_schema=auto_refetch_schema,
            connect_timeout=connect_timeout,
            reconnect_timeout=reconnect_timeout,
            request_timeout=request_timeout,
            ping_timeout=ping_timeout,
            encoding=encoding,
            initial_read_buffer_size=initial_read_buffer_size)
        await self._conn.connect()
        return self._conn

    async def tnt_disconnect(self):
        if hasattr(self, 'conn') and self.conn is not None:
            await self._conn.disconnect()
            self._conn = None

    async def tnt_reconnect(self, **kwargs):
        await self.tnt_disconnect()
        await self.tnt_connect(**kwargs)

    def assertResponseEqual(self, resp, target, *args):
        tuples = []
        for item in resp:
            if isinstance(item, TarantoolTuple):
                item = list(item)
            tuples.append(item)
        return self.assertListEqual(tuples, target, *args)

    def assertResponseEqualKV(self, resp, target, *args):
        tuples = []
        for item in resp:
            if isinstance(item, TarantoolTuple):
                item = dict(item)
            tuples.append(item)
        return self.assertListEqual(tuples, target, *args)


def ensure_version(*, min=None, max=None,
                   min_included=True, max_included=False):
    def check_version_wrap(f):
        @functools.wraps(f)
        async def wrap(self, *args, **kwargs):
            if check_version(self, self.conn.version,
                             min=min, max=max,
                             min_included=min_included,
                             max_included=max_included):
                res = f(self, *args, **kwargs)
                if inspect.isawaitable(res):
                    return await res
                return res

        return wrap

    return check_version_wrap


def ensure_bin_version(*, min=None, max=None,
                       min_included=True, max_included=False):
    def check_version_wrap(cls):
        assert issubclass(cls, TarantoolTestCase)
        cls.init_instance()
        ok, reason = _check_version(cls.tnt.bin_version,
                                    min=min, max=max,
                                    min_included=min_included,
                                    max_included=max_included)

        if not ok:
            cls = unittest.skip(reason)(cls)

        return cls

    return check_version_wrap


def check_version(test, version, *, min=None, max=None,
                  min_included=False, max_included=False):
    ok, reason = _check_version(version, min=min, max=max,
                                min_included=min_included,
                                max_included=max_included)

    if not ok:
        test.skipTest(reason)
        return False

    return True


def _check_version(version, *, min=None, max=None,
                   min_included=False,
                   max_included=False) -> Tuple[bool, Optional[str]]:
    if min and (version < min or (min_included and version <= min)):
        return False, f'version mismatch - required min={min} got={version}'

    if max and (version > max or (max_included and version >= max)):
        return False, f'version mismatch - required max={max} got={version}'

    return True, None



import datetime
import sys
import uuid
from decimal import Decimal

import dateutil.parser
import pytz

from asynctnt import IProtoError
from asynctnt._testbase import ensure_version
from asynctnt.exceptions import TarantoolDatabaseError, ErrorCode
from tests import BaseTarantoolTestCase


class MpExtTestCase(BaseTarantoolTestCase):

    @ensure_version(min=(2, 2))
    async def test__decimal(self):
        space = 'tester_ext_dec'

        dec = Decimal('-12.34')
        res = await self.conn.replace(space, [1, dec])
        self.assertEqual(res[0][1], dec)

        dec = Decimal('-12.345')
        res = await self.conn.replace(space, [1, dec])
        self.assertEqual(res[0][1], dec)

        dec = Decimal('-12.4')
        res = await self.conn.replace(space, [1, dec])
        self.assertEqual(res[0][1], dec)

        dec = Decimal('0')
        res = await self.conn.replace(space, [1, dec])
        self.assertEqual(res[0][1], dec)

        dec = Decimal('0.000')
        res = await self.conn.replace(space, [1, dec])
        self.assertEqual(res[0][1], dec)

        dec = Decimal(42)
        res = await self.conn.replace(space, [1, dec])
        self.assertEqual(res[0][1], dec)

        dec = Decimal('0.33')
        res = await self.conn.replace(space, [1, dec])
        self.assertEqual(res[0][1], dec)

        dec = Decimal('0.000000000000000000000000000000000010')
        res = await self.conn.replace(space, [1, dec])
        self.assertEqual(res[0][1], dec)

        dec = Decimal('-0.000000000000000000000000000000000010')
        res = await self.conn.replace(space, [1, dec])
        self.assertEqual(res[0][1], dec)

        dec = Decimal('0.1111111111111111')
        res = await self.conn.replace(space, [1, dec])
        self.assertEqual(res[0][1], dec)

        dec = Decimal('-0.1111111111111111')
        res = await self.conn.replace(space, [1, dec])
        self.assertEqual(res[0][1], dec)

        dec = Decimal('-0.111111')
        res = await self.conn.replace(space, [1, dec])
        self.assertEqual(res[0][1], dec)

    @ensure_version(min=(2, 4, 1))
    async def test__uuid(self):
        space = 'tester_ext_uuid'

        val = uuid.uuid4()
        res = await self.conn.replace(space, [1, val])
        self.assertEqual(res[0][1], val)

        val = uuid.UUID('f6423bdf-b49e-4913-b361-0740c9702e4b')
        res = await self.conn.replace(space, [1, val])
        self.assertEqual(res[0][1], val)

        val = uuid.UUID('00000000-0000-0000-0000-000000000000')
        res = await self.conn.replace(space, [1, val])
        self.assertEqual(res[0][1], val)

        val = uuid.uuid1(1, 100)
        res = await self.conn.replace(space, [1, val])
        self.assertEqual(res[0][1], val)

        val = uuid.uuid3(uuid.uuid4(), "hellothere")
        res = await self.conn.replace(space, [1, val])
        self.assertEqual(res[0][1], val)

        val = uuid.uuid5(uuid.NAMESPACE_URL, "generalkenobi")
        res = await self.conn.replace(space, [1, val])
        self.assertEqual(res[0][1], val)

    @ensure_version(min=(2, 4, 1))
    async def test__ext_error(self):
        try:
            await self.conn.eval("""
                box.schema.space.create('_space')
            """)
        except TarantoolDatabaseError as e:
            self.assertIsNotNone(e.error)
            self.assertGreater(len(e.error.trace), 0)
            frame = e.error.trace[0]
            self.assertEqual('ClientError', frame.error_type)
            self.assertIsNotNone(frame.file)
            self.assertIsNotNone(frame.line)
            self.assertEqual("Space '_space' already exists", frame.message)
            self.assertEqual(0, frame.err_no)
            self.assertEqual(ErrorCode.ER_SPACE_EXISTS, frame.code)

    @ensure_version(min=(2, 4, 1))
    async def test__ext_error_custom(self):
        try:
            await self.conn.eval("""
                local e = box.error.new{code=5,reason='A',type='B'}
                box.error(e)
            """)
        except TarantoolDatabaseError as e:
            self.assertIsNotNone(e.error)
            self.assertGreater(len(e.error.trace), 0)
            frame = e.error.trace[0]
            self.assertEqual('CustomError', frame.error_type)
            self.assertIsNotNone(frame.file)
            self.assertIsNotNone(frame.line)
            self.assertEqual("A", frame.message)
            self.assertEqual(0, frame.err_no)
            self.assertEqual(5, frame.code)
            self.assertIn('custom_type', frame.fields)
            self.assertEqual('B', frame.fields['custom_type'])

    @ensure_version(min=(2, 10))
    async def test__ext_error_custom_return(self):
        resp = await self.conn.eval("""
            local e = box.error.new{code=5,reason='A',type='B'}
            return e
        """)
        e = resp[0]
        self.assertIsInstance(e, IProtoError)
        self.assertGreater(len(e.trace), 0)
        frame = e.trace[0]
        self.assertEqual('CustomError', frame.error_type)
        self.assertEqual('eval', frame.file)
        self.assertEqual(2, frame.line)
        self.assertEqual("A", frame.message)
        self.assertEqual(0, frame.err_no)
        self.assertEqual(5, frame.code)
        self.assertIn('custom_type', frame.fields)
        self.assertEqual('B', frame.fields['custom_type'])

    @ensure_version(min=(2, 10))
    async def test__ext_error_custom_return_with_disabled_exterror(self):
        await self.conn.eval("""
            require('msgpack').cfg{encode_error_as_ext = false}
        """)
        try:
            resp = await self.conn.eval("""
                local e = box.error.new{code=5,reason='A',type='B'}
                return e
            """)
            e = resp[0]
            self.assertIsInstance(e, str)
            self.assertEqual('A', e)
        finally:
            await self.conn.eval("""
                require('msgpack').cfg{encode_error_as_ext = true}
            """)

    @ensure_version(min=(2, 10))
    async def test__ext_datetime_read(self):
        resp = await self.conn.eval("""
            local date = require('datetime')
            return date.parse('2000-01-01T02:00:00.23+0300')
        """)
        res = resp[0]
        dt = datetime_fromisoformat('2000-01-01T02:00:00.230000+03:00')
        self.assertEqual(dt, res)

    @ensure_version(min=(2, 10))
    async def test__ext_datetime_tz(self):
        resp = await self.conn.eval("""
            local date = require('datetime')
            return date.parse('2000-01-01T02:00:00 MSK')
        """)
        res = resp[0]
        dt = datetime_fromisoformat('2000-01-01T02:00:00+03:00')
        self.assertEqual(dt, res)

    @ensure_version(min=(2, 10))
    async def test__ext_datetime_read_neg_tz(self):
        resp = await self.conn.eval("""
            local date = require('datetime')
            return date.parse('2000-01-01T02:17:43.23-08:00')
        """)
        res = resp[0]
        dt = datetime_fromisoformat('2000-01-01T02:17:43.230000-08:00')
        self.assertEqual(dt, res)

    @ensure_version(min=(2, 10))
    async def test__ext_datetime_read_before_1970(self):
        resp = await self.conn.eval("""
            local date = require('datetime')
            return date.parse('1930-01-01T02:17:43.23-08:00')
        """)
        res = resp[0]
        dt = datetime_fromisoformat('1930-01-01T02:17:43.230000-08:00')
        self.assertEqual(dt, res)

    @ensure_version(min=(2, 10))
    async def test__ext_datetime_write(self):
        sp = 'tester_ext_datetime'
        dt = datetime_fromisoformat('2000-01-01T02:17:43.230000-08:00')
        resp = await self.conn.insert(sp, [1, dt])
        res = resp[0]
        self.assertEqual(dt, res['dt'])

    @ensure_version(min=(2, 10))
    async def test__ext_datetime_write_before_1970(self):
        sp = 'tester_ext_datetime'
        dt = datetime_fromisoformat('1004-01-01T02:17:43.230000+04:00')
        resp = await self.conn.insert(sp, [1, dt])
        res = resp[0]
        self.assertEqual(dt, res['dt'])

    @ensure_version(min=(2, 10))
    async def test__ext_datetime_write_without_tz(self):
        sp = 'tester_ext_datetime'
        dt = datetime_fromisoformat('2022-04-23T02:17:43.450000')
        resp = await self.conn.insert(sp, [1, dt])
        res = resp[0]
        self.assertEqual(dt, res['dt'])

    @ensure_version(min=(2, 10))
    async def test__ext_datetime_write_without_tz_integer(self):
        sp = 'tester_ext_datetime'
        dt = datetime_fromisoformat('2022-04-23T02:17:43')
        resp = await self.conn.insert(sp, [1, dt])
        res = resp[0]
        self.assertEqual(dt, res['dt'])

    @ensure_version(min=(2, 10))
    async def test__ext_datetime_write_pytz(self):
        sp = 'tester_ext_datetime'
        dt = datetime_fromisoformat('2022-04-23T02:17:43')
        dt = pytz.timezone('Europe/Amsterdam').localize(dt)
        resp = await self.conn.insert(sp, [1, dt])
        res = resp[0]
        self.assertEqual(dt, res['dt'])

    @ensure_version(min=(2, 10))
    async def test__ext_datetime_write_pytz_america(self):
        sp = 'tester_ext_datetime'
        dt = datetime_fromisoformat('2022-04-23T02:17:43')
        dt = pytz.timezone('America/New_York').localize(dt)
        resp = await self.conn.insert(sp, [1, dt])
        res = resp[0]
        self.assertEqual(dt, res['dt'])


def datetime_fromisoformat(s):
    if sys.version_info < (3, 7, 0):
        return dateutil.parser.isoparse(s)
    return datetime.datetime.fromisoformat(s)
